import logging
import os
import sys

def get_loglevel() -> int:
    """
    returns log level 
        if LOG_LEVEL is in environment variables
    or
        'INFO'
    """
    _default_level = 'INFO'
    LOG_LEVEL = os.environ.get('LOG_LEVEL', _default_level).upper()
    if LOG_LEVEL == '':
        LOG_LEVEL = _default_level

    return logging._nameToLevel[LOG_LEVEL]

LOG_LEVEL = get_loglevel()

def get_value_or_default(key:str, default:any, **kwargs):
    """
    returns kwargs value or default
        if key is in kwargs
    or
        default value
    """
    return kwargs[key] if key in kwargs else default


def load_test_logger(level = logging.INFO):
  test_logger = logging.getLogger()
  test_logger.setLevel(level=level)
  console_handler = logging.StreamHandler(stream=sys.stdout)
  console_handler.formatter = logging.Formatter('::> %(name)s - %(levelname)s - %(message)s')
  test_logger.handlers.clear()
  test_logger.addHandler(console_handler)


import logging
import inspect
from functools import wraps
from typing import Callable, TypeVar, Any, cast
from random import randrange

# Configure logging (adjust as needed)
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)

F = TypeVar("F", bound=Callable[..., Any])


def logged(
    message: str,
    level: int = logging.INFO,
    show_in: bool = True,
    show_out: bool = True,
    show_data: bool = True,
) -> Callable[[F], F]:
    """
    Decorator to log the start and finish of a function execution.

    Args:
        message (str): Custom message to include in logs.
        level (int): Logging level.
        show_in (bool): Whether to log the start of the function.
        show_out (bool): Whether to log the finish of the function.
        show_data (bool): Whether to log the args and kwargs of the function.

    Returns:
        Callable: The decorated function with logging.
    """

    def decorator(func: F) -> F:
        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def awrapper(*args, **kwargs) -> Any:
                id = randrange(1, 100_000_000_000)
                try:
                    if show_in:
                        logging.log(
                            level,
                            f"{message} - Starting({id}) *{func.__name__}* with args: {args if show_data else ''}, kwargs: {kwargs if show_data else ''}",
                        )

                    result = await func(*args, **kwargs)

                    if show_out:
                        logging.log(
                            level,
                            f"{message} - Finished({id}) *{func.__name__}* with result: {result if show_data else ''}",
                        )
                    return result
                except Exception as e:
                    logging.exception(
                        f"{message} - Exception({id}) *{func.__name__}* raised an exception: {e}"
                    )
                    raise  # Re-raise the exception after logging

            return cast(F, awrapper)
        else:

            @wraps(func)
            def wrapper(*args, **kwargs) -> Any:
                id = randrange(1, 100_000_000_000)
                try:
                    if show_in:
                        logging.log(
                            level,
                            f"{message} - Starting({id}) *{func.__name__}* with args: {args if show_data else ''}, kwargs: {kwargs if show_data else ''}",
                        )

                    result = func(*args, **kwargs)

                    if show_out:
                        logging.log(
                            level,
                            f"{message} - Finished({id}) *{func.__name__}* with result: {result if show_data else ''}",
                        )
                    return result
                except Exception as e:
                    logging.exception(
                        f"{message} - Exception({id}) *{func.__name__}* raised an exception: {e}"
                    )
                    raise

            return cast(F, wrapper)

    return decorator



from contextlib import contextmanager, asynccontextmanager
import logging
from typing import Optional, Generator, AsyncGenerator


@contextmanager
def logged_step(
    name: str, level: int = logging.INFO, raise_exception: bool = True
) -> Generator[None, None, None]:
    """
    Context manager to log the start and finish of a step, and handle exceptions.

    Args:
        name (str): Name of the step to log.
        level (int, optional): Logging level. Defaults to logging.INFO.
        raise_exception (bool, optional): Whether to re-raise exceptions after logging. Defaults to True.

    Yields:
        None
    """
    try:
        logging.log(level, f"Step({name}): Executing...")
        yield
        logging.log(level, f"Step({name}): Executed.")
    except Exception as ex:
        # Log the exception at ERROR level regardless of the specified level
        logging.error(f"Step({name}): Failed! Reason: {ex}")
        if raise_exception:
            raise


@asynccontextmanager
async def alogged_step(
    name: str, level: int = logging.INFO, raise_exception: bool = True
) -> AsyncGenerator[None, None]:
    """
    Asynchronous context manager to log the start and finish of a step, and handle exceptions.

    Args:
        name (str): Name of the step to log.
        level (int, optional): Logging level. Defaults to logging.INFO.
        raise_exception (bool, optional): Whether to re-raise exceptions after logging. Defaults to True.

    Yields:
        None
    """
    try:
        logging.log(level, f"Step({name}): Executing...")
        yield
        logging.log(level, f"Step({name}): Executed.")
    except Exception as ex:
        # Log the exception at ERROR level regardless of the specified level
        logging.error(f"Step({name}): Failed! Reason: {ex}")
        if raise_exception:
            raise



from . import utils
from .settings.Settings import Settings
from . import errors
from . import decorators
from . import context_managers

__all__ = ["utils", "Settings", "errors", "decorators", "context_managers"]



import logging
import inspect
from functools import wraps
from typing import Optional
from time import perf_counter
from typing import Optional, Callable, TypeVar, Any

F = TypeVar("F", bound=Callable[..., Any])


def timed(
    _func: Optional[F] = None,
    *,
    message: Optional[str] = None,
    level: int = logging.INFO,
) -> Callable[..., F]:
    """
    A decorator that logs the time taken by a function to execute.
    Can be used with or without arguments.

    Usage:
        @timed
        def func(...):
            ...

        @timed(message="Custom message", level=logging.DEBUG)
        def func(...):
            ...
    """

    def decorator(func: F) -> F:
        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapper(*args, **kwargs) -> Any:
                start = perf_counter()
                try:
                    result = await func(*args, **kwargs)
                    return result
                finally:
                    elapsed = perf_counter() - start
                    log_message = (
                        f"{func.__name__} took {elapsed:.6f} seconds to finish"
                    )
                    if message:
                        log_message = f"{message} - {log_message}"
                    logging.log(level, f"Timed: {log_message}")

            return async_wrapper  # type: ignore
        else:

            @wraps(func)
            def sync_wrapper(*args, **kwargs) -> Any:
                start = perf_counter()
                try:
                    result = func(*args, **kwargs)
                    return result
                finally:
                    elapsed = perf_counter() - start
                    log_message = (
                        f"{func.__name__} took {elapsed:.6f} seconds to finish"
                    )
                    if message:
                        log_message = f"{message} - {log_message}"
                    logging.log(level, f"Timed: {log_message}")

            return sync_wrapper  # type: ignore

    if _func is None:
        # Decorator is called with arguments
        return decorator
    else:
        # Decorator is used without arguments
        return decorator(_func)



import yaml
import logging
from typing import List, Literal
from ..errors import YamlParseError
import logging

logging.basicConfig(format="%(asctime)s - %(levelname)s - %(name)s - %(message)s")
logger = logging.getLogger(__name__)
import re
import os


class Settings(dict):
    """
    Represents app settings yaml file
    """

    def __init__(
        self,
        section: Literal["llm", "embeddings", "vectordb"],
        path="./app-settings.yaml",
        **kwargs,
    ):
        """
        Reads given yaml file
        - Use kwargs to override values
        """
        self.path = path
        self.section = section

        try:
            # read yaml file
            yaml_content = self._read_yaml(path)

            self._replace_with_environment_variables(yaml_content)

            # add values
            yaml_section = yaml_content[section]
            for key, value in yaml_section.items():
                self[key] = value

            # override values
            self.update({key: value for key, value in kwargs.items()})

        except Exception as ex:
            logger.error(
                f"Failed to read '{section}' section from '{self.path}'. Original: {ex}"
            )
            raise

    def _read_yaml(self, path):
        try:
            with open(path, "r") as file:
                yaml_content = yaml.safe_load(file)
                return yaml_content
        except FileNotFoundError:
            raise
        except Exception as ex:
            raise YamlParseError(
                f"File '{self.path}' could not be parsed correcty to yaml! Original: {ex}"
            )

    def _set_nested_value(self, yaml_data: dict, keys: List[str], value):
        for key in keys[:-1]:
            yaml_data = yaml_data.setdefault(key, {})
        if value is not None and value:
            yaml_data[keys[-1]] = value

    def _replace_with_environment_variables(self, yaml_content):
        # Pattern to match environment variables with double underscores
        pattern = re.compile(r"(?i)^[a-z][a-z0-9_]*(?:__[a-z][a-z0-9_]*)+$")

        # Process each environment variable
        for key, value in os.environ.items():
            if pattern.match(key):
                # Split the key into parts
                keys = key.lower().split("__")
                # Update the YAML configuration
                self._set_nested_value(yaml_content, keys, value)

    def __repr__(self) -> str:
        return self._hide_keys().__repr__()

    def __str__(self) -> str:
        return self._hide_keys().__str__()

    def _hide_keys(self) -> dict:
        temp = super().copy()
        pairs = [(key, value) for key, value in temp.items() if "key" in key]
        for key, value in pairs:
            value_str = str(value)
            if len(value_str) < 7:
                temp[key] = "*******"
            else:
                temp[key] = f"{value_str[:2]}*****{value_str[-2:]}"
        return temp
